/*
 * GnuAPDF Library
 *   Jon Sevy, jsevy@cs.drexel.edu
 * 
 * Based on GnuJPDF Library, authored by (among others)
 *   Peter T Mount, http://www.retep.org.uk/pdf/
 *   Eric Z. Beard, ericzbeard@hotmail.com
 *   Gilbert DeLeeuw, gil1@users.sourceforge.net
 *   Matthew Hreljac, mhreljac@hotmail.com  
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package gnu.apdftest;


import gnu.apdf.PDFCanvas;
import gnu.apdf.PDFDocument;
import gnu.apdf.PDFPage;
import gnu.apdf.PDFPageFormat;
import gnu.apdftest.R;

import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import android.annotation.SuppressLint;
import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Matrix;
import android.graphics.Paint;
import android.graphics.Paint.Style;
import android.graphics.Path;
import android.graphics.Picture;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.Typeface;
import android.os.Environment;
import android.util.AttributeSet;
import android.view.View;


public class PDFTestView extends View
{
        
    public enum DisplayContent
    {
    	GRAPHICS,
    	TEXT,
    	TRANSFORMATIONS,
    	CLIPPING,
    	IMAGES,
    	BITMAP_MESH,
    	VERTICES_TRIANGLES,
    	VERTICES_FAN,
    	VERTICES_STRIP
    }
    
    private Map<Integer, String> fileSuffixMap = new HashMap<Integer, String>()
    {
        {
            put(R.id.testGraphicsButton, "graphics");
            put(R.id.testTextButton, "text");
            put(R.id.testClippingButton, "clipping");
            put(R.id.testTransformationsButton, "transformations");
            put(R.id.testImagesButton, "images");
            put(R.id.testBitmapMeshButton, "mesh");
            put(R.id.testDrawVerticesTraingleButton, "vertices");
            put(R.id.testDrawVerticeFanButton, "vertices_fan");
            put(R.id.testDrawVerticesStripButton, "vertices_strip");
        }
    };
    
    
    // what to display and generate in the PDF file - just use button ID
    private int displayContent = R.id.testGraphicsButton;
    
    
    public void setDisplayContent(int displayContentButtonId)
    {
    	this.displayContent = displayContentButtonId;
    }
    
    
    
    public PDFTestView(Context context, AttributeSet attrs)
    {
        super(context, attrs);
    }
    
    
    
    public void onDraw(Canvas canvas)
    {
        drawDesign(canvas);
    }
    
    
    
    public void generatePDF()
    {
        PDFDocument pdfDoc = new PDFDocument();
        PDFPage page = new PDFPage(PDFPageFormat.A0);
        pdfDoc.add(page);
        PDFCanvas pdfCanvas = page.getCanvas();
        
        drawDesign(pdfCanvas);
        
        ByteArrayOutputStream outStream = new ByteArrayOutputStream();
        try
        {
            pdfDoc.write(outStream);
            
            // print pdf output to console for convenient inspection
            String stringOutput = outStream.toString();
            System.out.println(stringOutput);
            
            // now write to files - one with .pdf for display, and one with .txt for inspection
            // get appropriate suffix depnding on what's generated
            String fileSuffix = fileSuffixMap.get(displayContent);
            if (fileSuffix == null)
            {
            	// hmm... not sure why we didn't find it - just leave off
            	fileSuffix = "";
            }
            
            String pdfPath = Environment.getExternalStoragePublicDirectory(Environment.DIRECTORY_DOWNLOADS).getAbsolutePath() + "/test_pdf_file_generation_" + fileSuffix + ".pdf";
            // FileWriter fileWriter = new FileWriter(pdfPath);
            // fileWriter.write(output);
            // fileWriter.flush();
            // fileWriter.close();
            BufferedOutputStream fileOutputStream = new BufferedOutputStream(new FileOutputStream(pdfPath));
            fileOutputStream.write(outStream.toByteArray());
            fileOutputStream.flush();
            fileOutputStream.close();
            
            // write a text file (*.txt) so can check out the generated pdf
            String textPath = Environment.getExternalStoragePublicDirectory(Environment.DIRECTORY_DOWNLOADS).getAbsolutePath() + "/test_pdf_file_generation_" + fileSuffix + ".txt";
            FileWriter textFileWriter = new FileWriter(textPath);
            textFileWriter.write(stringOutput);
            textFileWriter.flush();
            textFileWriter.close();
        }
        catch (IOException e)
        {
            System.out.println("Bah!!!!! " + e.toString());
        }
    }
    
    
    
    /**
     * Used to draw both on the screen and to the PDF document, to check if the PDF image matches the
     * one on the screen
     * @param canvas  Either a "normal" canvas for screen rendering or one associated with a PDF Page
     */
    private void drawDesign(Canvas canvas)
    {
        // draw some stuff
        Paint paint = new Paint();
        
        // fill the page with a pale yellow
        paint.setColor(Color.rgb(255, 245, 240));
        paint.setStyle(Style.FILL);
        canvas.drawPaint(paint);
        
        switch (displayContent)
        {
        
	        case R.id.testGraphicsButton:
	        {
	        	testGraphics(canvas);
	        	break;
	        }
	        
	        case R.id.testTextButton:
	        {
	        	testText(canvas);
	        	break;
	        }
	        
	        case R.id.testClippingButton:
	        {
	            testClipping(canvas);
	        	break;
	        }
	        case R.id.testTransformationsButton:
	        {
	            testTransformations(canvas);
	        	break;
	        }
	        case R.id.testImagesButton:
	        {
	            testImages(canvas);
	        	break;
	        }
	        case R.id.testBitmapMeshButton:
	        {
	            testGrid(canvas);
	        	break;
	        }
	        
	        case R.id.testDrawVerticesTraingleButton:
	        {
	            testDrawVerticesTriangleMode(canvas);
	        	break;
	        }
	        case R.id.testDrawVerticeFanButton:
	        {
	            testDrawVerticesTriangleFanMode(canvas);
	        	break;
	        }
	        case R.id.testDrawVerticesStripButton:
	        {
	            testDrawVerticesTriangleStripMode(canvas);
	        	break;
	        }
	        
	        default:
	        {
	        	testGraphics(canvas);
	        	break;
	        }
	        
        
        };
        
    }
    
    
    
    private void testImages(Canvas canvas)
    {
        Paint paint = new Paint();
        paint.setColor(Color.BLUE);
        paint.setStyle(Style.STROKE);
        
        // draw a bitmap
        // create bitmap from embedded image
        Bitmap bitmap = BitmapFactory.decodeResource(getContext().getResources(), R.drawable.skull);
        
        // draw at given size
        float x = 100;
        float y = 200;
        float width = bitmap.getWidth();
        float height = bitmap.getHeight();
        canvas.drawBitmap(bitmap, x, y, paint);
        canvas.drawRect(x, y, x + width, y + height, paint);
        
        // draw enlarged and stretched
        Rect srcRect = new Rect(0, 0, bitmap.getWidth(), bitmap.getHeight());
        RectF dstRect = new RectF(300, 300, 600, 400);
        canvas.drawBitmap(bitmap, srcRect, dstRect, paint);
        canvas.drawRect(dstRect, paint);
        
        // draw just a subset of the image
        srcRect = new Rect(bitmap.getWidth() / 4, 0, 3 * bitmap.getWidth() / 4, bitmap.getHeight() / 2);
        dstRect = new RectF(100, 500, 100 + bitmap.getWidth() / 2, 500 + bitmap.getHeight() / 2);
        canvas.drawBitmap(bitmap, srcRect, dstRect, paint);
        canvas.drawRect(dstRect, paint);
        
        // use transformation and see what we get
        x = 50;
        y = 300;
        for (int i = 0; i < 90; i += 10)
        {
            canvas.rotate(-i);
            canvas.scale(2f, 3f);
            canvas.drawBitmap(bitmap, x, y, paint);
            canvas.scale(1 / 2f, 1 / 3f);
            canvas.rotate(i);
        }
        
        canvas.translate(200, 200);
        canvas.skew(0.5f, 0.1f);
        canvas.rotate(20);
        
        // now supply transformation and see what happens
        Matrix matrix = new Matrix();
        matrix.preScale(3, 3);
        matrix.postRotate(-10);
        matrix.postTranslate(200, 400);
        canvas.drawBitmap(bitmap, matrix, paint);
        
        /*
         * // try drawing a Picture canvas.drawPicture(picture); try { picture.draw(canvas); } catch (Exception e) { System.out.println("Failed to draw picture: " + e.toString()); }
         */
    }
    
    
    /*
     * 
     * Alas, drawPicture doesn't work with PDFCanvas - the saved picture doesn't actually
     * "re-roll" the drawing commands, but does something natively...
    @SuppressLint("NewApi")
    private void testPicture(Canvas canvas)
    {
        // create a picture
        Picture picture;
        
        Paint paint = new Paint();
        paint.setStyle(Style.STROKE);
        paint.setColor(Color.RED);
        paint.setStrokeWidth(3);
        
        // need to disable hardware acceleration, or drawPicture doesn't work
        setLayerType(View.LAYER_TYPE_SOFTWARE, null);
        
        picture = new Picture();
        Canvas recordingCanvas = picture.beginRecording(500, 500);
        
        //recordingCanvas.drawColor(Color.LTGRAY);
        recordingCanvas.drawCircle(100, 100, 80, paint);
        recordingCanvas.drawRect(50, 50, 400, 300, paint);
        recordingCanvas.drawLine(250, 150, 450, 350, paint);
        
        picture.endRecording();
        
        // try drawing a Picture 
        canvas.drawPicture(picture); 
        
    }
    */
    
    
    
    private void testGrid(Canvas canvas)
    {
        Paint paint = new Paint();
        
        // draw stuff in blue
        paint.setColor(Color.BLUE);
        paint.setStyle(Style.STROKE);
        
        /*
         * Test of drawBitmapMesh function - stretches a bitmap across a square mesh
         */
        
        // create a 3D-looking mesh by projecting a surface onto the plane
        // use a simple paraboloid z = x^2 + y^2
        int gridCountX = 10;
        int gridCountY = 10;
        float xLeft = 2;
        float xRight = -2;
        float xDelta = (xRight - xLeft) / gridCountX;
        float yLeft = 2;
        float yRight = -2;
        float yDelta = (yRight - yLeft) / gridCountY;
        
        Vector3D[][] vectors = new Vector3D[gridCountX + 1][gridCountY + 1];
        
        // create surface points
        for (int i = 0; i <= gridCountX; i++)
        {
            for (int j = 0; j <= gridCountY; j++)
            {
                vectors[i][j] = new Vector3D();
                vectors[i][j].x = xLeft + i * xDelta;
                vectors[i][j].y = yLeft + j * yDelta;
                vectors[i][j].z = (float) Math.pow(vectors[i][j].x, 2) + (float) Math.pow(vectors[i][j].y, 2);
            }
        }
        
        // now project these onto plane, using view vector (1, 2, 1) and "up"
        // vector (0, 0, 1)
        Vector3D pov = new Vector3D(1, 1, 10);
        Vector3D up = new Vector3D(0, 1, 0);
        
        // first basis vector is up minus its projection on pov: b1 = up - (up
        // dot pov)*pov, normalized
        // Vector3D b1 = up.sum(pov.scalarProduct(-up.dotProduct(pov)));
        pov = pov.scalarProduct(1 / pov.magnitude());
        
        float dot = -up.dotProduct(pov);
        Vector3D proj = pov.scalarProduct(dot);
        Vector3D b1 = up.sum(proj);
        b1 = b1.scalarProduct(1 / b1.magnitude());
        
        // second basis vector is cross product up cross pov, normalized
        Vector3D b2 = up.crossProduct(pov);
        b2 = b2.scalarProduct(1 / b2.magnitude());
        
        /*
         * float[][][] coords = new float[gridCount + 1][gridCount + 1][2];
         * 
         * for (int i = 0; i <= gridCount; i++) { for (int j = 0; j <= gridCount; j++) { coords[i][j][0] = vectors[i][j].dotProduct(b2); coords[i][j][1] = vectors[i][j].dotProduct(b1); } }
         * 
         * // draw the grid...but first translate and scale canvas.scale(200, 200); canvas.translate(4, 4);
         * 
         * //paint.setStrokeWidth(1);
         * 
         * for (int i = 0; i <= gridCount; i++) { for (int j = 0; j <= gridCount; j++) { // draw lines from vertex (i, j) to its right and lower neighbors if (i < gridCount)
         * canvas.drawLine(coords[i][j][0], coords[i][j][1], coords[i+1][j][0], coords[i+1][j][1], paint); if (j < gridCount) canvas.drawLine(coords[i][j][0], coords[i][j][1], coords[i][j+1][0],
         * coords[i][j+1][1], paint); } }
         */
        
        // now the new x and y coords are just the dot products of the vectors
        // above with these
        
        // the Canvas routine requires these in a single-dimension array as x, y
        // pairs
        float[] coords = new float[2 * (gridCountX + 1) * (gridCountY + 1)];
        for (int i = 0; i <= gridCountX; i++)
        {
            for (int j = 0; j <= gridCountY; j++)
            {
                int posX = 2 * (i + j * (gridCountX + 1));
                int posY = posX + 1;
                coords[posX] = vectors[i][j].dotProduct(b2);
                coords[posY] = vectors[i][j].dotProduct(b1);
            }
        }
        
        // now draw these on the canvas... but first translate and scale
        canvas.scale(100, 100);
        canvas.translate(3, 3);
        // paint.setStrokeWidth(1);
        
        /*
         * paint.setColor(Color.BLUE); canvas.drawLine(-6, -6, coords[0], coords[1], paint); paint.setColor(Color.RED); canvas.drawLine(-6, -6, coords[2], coords[3], paint);
         * paint.setColor(Color.GREEN); canvas.drawLine(-6, -6, coords[4], coords[5], paint); paint.setColor(Color.YELLOW); canvas.drawLine(-6, -6, coords[6], coords[7], paint);
         * paint.setColor(Color.BLUE); canvas.drawLine(coords[0], coords[1], coords[2], coords[3], paint); canvas.drawLine(coords[0], coords[1], coords[4], coords[5], paint);
         * canvas.drawLine(coords[6], coords[7], coords[4], coords[5], paint); canvas.drawLine(coords[6], coords[7], coords[2], coords[3], paint);
         */
        
        // create bitmap from embedded image
        Bitmap bitmap = BitmapFactory.decodeResource(getContext().getResources(), R.drawable.skull);
        
        // draw the bitmap on the grid
        canvas.drawBitmapMesh(bitmap, gridCountX, gridCountY, coords, 0, null, 0, paint);
        
        // draw the grid...
        for (int i = 0; i <= gridCountX; i++)
        {
            for (int j = 0; j <= gridCountY; j++)
            {
                int centerX = 2 * (i + j * (gridCountX + 1));
                int rightX = 2 * ((i + 1) + j * (gridCountX + 1));
                int lowerX = 2 * (i + (j + 1) * (gridCountX + 1));
                
                int centerY = centerX + 1;
                int rightY = rightX + 1;
                int lowerY = lowerX + 1;
                
                // draw lines from vertex (i, j) to its right and lower
                // neighbors
                if (i < gridCountX)
                    canvas.drawLine(coords[centerX], coords[centerY], coords[rightX], coords[rightY], paint);
                if (j < gridCountY)
                    canvas.drawLine(coords[centerX], coords[centerY], coords[lowerX], coords[lowerY], paint);
            }
        }
        
        /*
         * float[] coords = {0, 0, 400, 0, 200, 800, 600, 600}; Matrix matrix = new Matrix(); matrix.preRotate(-20); matrix.preTranslate(100, 200); matrix.mapPoints(coords);
         * 
         * canvas.translate(0, 400); canvas.rotate(15);
         * 
         * int gridCount = 1;
         * 
         * paint.setStrokeWidth(1); paint.setColor(Color.BLUE); canvas.drawCircle(coords[0], coords[1], 20, paint); paint.setColor(Color.RED); canvas.drawCircle(coords[2], coords[3], 20, paint);
         * paint.setColor(Color.GREEN); canvas.drawCircle(coords[4], coords[5], 20, paint); paint.setColor(Color.YELLOW); canvas.drawCircle(coords[6], coords[7], 20, paint);
         * paint.setColor(Color.BLUE); canvas.drawLine(coords[0], coords[1], coords[2], coords[3], paint); canvas.drawLine(coords[0], coords[1], coords[4], coords[5], paint);
         * canvas.drawLine(coords[6], coords[7], coords[4], coords[5], paint); canvas.drawLine(coords[6], coords[7], coords[2], coords[3], paint);
         * 
         * // create bitmap from embedded image Bitmap bitmap = BitmapFactory.decodeResource(getContext().getResources(), R.drawable.skull);
         * 
         * // draw the bitmap on the grid canvas.drawBitmapMesh(bitmap, gridCount, gridCount, coords, 0, null, 0, paint);
         */
    }
    
    
    
    @SuppressLint("NewApi")
    private void testDrawVerticesTriangleMode(Canvas canvas)
    {
        Paint paint = new Paint();
        
        // draw stuff in black
        paint.setColor(Color.BLACK);
        paint.setStyle(Style.STROKE);
        
        /*
         * Test of drawVertices function - draws a mesh of triangles, shaded by supplied color array
         */
        
        // create a 3D-looking mesh by projecting a surface onto the plane
        // use a simple paraboloid z = x^2 + y^2
        int gridCountX = 10;
        int gridCountY = 10;
        float xLeft = 2;
        float xRight = -2;
        float xDelta = (xRight - xLeft) / gridCountX;
        float yLeft = 2;
        float yRight = -2;
        float yDelta = (yRight - yLeft) / gridCountY;
        
        Vector3D[][] vectors = new Vector3D[gridCountX + 1][gridCountY + 1];
        
        // create surface points
        for (int i = 0; i <= gridCountX; i++)
        {
            for (int j = 0; j <= gridCountY; j++)
            {
                vectors[i][j] = new Vector3D();
                vectors[i][j].x = xLeft + i * xDelta;
                vectors[i][j].y = yLeft + j * yDelta;
                vectors[i][j].z = (float) Math.pow(vectors[i][j].x, 2) + (float) Math.pow(vectors[i][j].y, 2);
            }
        }
        
        // now project these onto plane, using view vector (1, 2, 1) and "up"
        // vector (0, 0, 1)
        Vector3D pov = new Vector3D(1, 1, 10);
        Vector3D up = new Vector3D(0, 1, 0);
        
        // first basis vector is up minus its projection on pov: b1 = up - (up
        // dot pov)*pov, normalized
        // Vector3D b1 = up.sum(pov.scalarProduct(-up.dotProduct(pov)));
        pov = pov.scalarProduct(1 / pov.magnitude());
        
        float dot = -up.dotProduct(pov);
        Vector3D proj = pov.scalarProduct(dot);
        Vector3D b1 = up.sum(proj);
        b1 = b1.scalarProduct(1 / b1.magnitude());
        
        // second basis vector is cross product up cross pov, normalized
        Vector3D b2 = up.crossProduct(pov);
        b2 = b2.scalarProduct(1 / b2.magnitude());
        
        // now the new x and y coords are just the dot products of the vectors
        // above with these basis vectors
        // the drawVertices routine requires that we specify a set of triangles
        // in a single-dimension array as x, y pair triples;
        // we have a total of twice the number of triangles as squares in the
        // grid, 2 coords per point, and 3 points per triangle
        float[] coords = new float[2 * (gridCountX) * (gridCountY) * 2 * 3];
        // also supply a color array for use with drawVertices
        int[] colors = new int[2 * (gridCountX) * (gridCountY) * 2 * 3];
        int pos = 0;
        for (int i = 0; i < gridCountX; i++)
        {
            for (int j = 0; j < gridCountY; j++)
            {
                // top triangle in grid first
                // top left point, color and then x and y
                colors[pos] = Color.rgb((int) (255 * i / (gridCountX + 1)), 0, (int) (255 * j / (gridCountY + 1)));
                coords[2 * pos] = vectors[i][j].dotProduct(b2);
                coords[2 * pos + 1] = vectors[i][j].dotProduct(b1);
                pos++;
                
                // top right point, color and then x and y
                colors[pos] = Color.rgb((int) (255 * (i + 1) / (gridCountX + 1)), 0, (int) (255 * j / (gridCountY + 1)));
                coords[2 * pos] = vectors[i + 1][j].dotProduct(b2);
                coords[2 * pos + 1] = vectors[i + 1][j].dotProduct(b1);
                pos++;
                
                // bottom right point, color and then x and y
                colors[pos] = Color.rgb((int) (255 * (i + 1) / (gridCountX + 1)), 0, (int) (255 * (j + 1) / (gridCountY + 1)));
                coords[2 * pos] = vectors[i + 1][j + 1].dotProduct(b2);
                coords[2 * pos + 1] = vectors[i + 1][j + 1].dotProduct(b1);
                pos++;
                
                // bottom triangle in grid next
                // top left point, color and then x and y
                colors[pos] = Color.rgb((int) (255 * i / (gridCountX + 1)), 0, (int) (255 * j / (gridCountY + 1)));
                coords[2 * pos] = vectors[i][j].dotProduct(b2);
                coords[2 * pos + 1] = vectors[i][j].dotProduct(b1);
                pos++;
                
                // bottom left point, color and then x and y
                colors[pos] = Color.rgb((int) (255 * i / (gridCountX + 1)), 0, (int) (255 * (j + 1) / (gridCountY + 1)));
                coords[2 * pos] = vectors[i][j + 1].dotProduct(b2);
                coords[2 * pos + 1] = vectors[i][j + 1].dotProduct(b1);
                pos++;
                
                // bottom right point, color and then x and y
                colors[pos] = Color.rgb((int) (255 * (i + 1) / (gridCountX + 1)), 0, (int) (255 * (j + 1) / (gridCountY + 1)));
                coords[2 * pos] = vectors[i + 1][j + 1].dotProduct(b2);
                coords[2 * pos + 1] = vectors[i + 1][j + 1].dotProduct(b1);
                pos++;
                
            }
        }
        
        // now draw these on the canvas... but first translate and scale
        canvas.scale(300, 300);
        canvas.translate(3, 3);
        // paint.setStrokeWidth(1);
        
        /*
         * coords[0] = 1; coords[1] = 1; coords[2] = 2; coords[3] = 1; coords[4] = 2; coords[5] = 3;
         * 
         * coords[6] = 1; coords[7] = 3; coords[8] = 0; coords[9] = 2; coords[10] = 0; coords[11] = 1;
         * 
         * colors[0] = Color.RED; colors[1] = Color.RED; colors[2] = Color.BLUE;
         * 
         * colors[3] = Color.BLUE; colors[4] = Color.WHITE; colors[5] = Color.WHITE;
         * 
         * colors[6] = Color.RED; colors[7] = Color.WHITE; colors[8] = Color.BLUE;
         * 
         * colors[9] = Color.RED; colors[10] = Color.WHITE; colors[11] = Color.BLUE;
         */
        
        // need to disable hardware acceleration, or drawVertices doesn't work
        setLayerType(View.LAYER_TYPE_SOFTWARE, null);
        canvas.drawVertices(Canvas.VertexMode.TRIANGLES, 2 * (gridCountX) * (gridCountY) * 2 * 3, coords, 0, null, 0, colors, 0, null, 0, 0, paint);
        // canvas.drawVertices(Canvas.VertexMode.TRIANGLE_FAN, 12, coords, 0,
        // null, 0, colors, 0, null, 0, 0, paint);
        
        // draw the grid over top for emphasis...
        for (int i = 0; i < coords.length; i += 6)
        // for (int i = 0; i < 12; i += 6)
        {
            canvas.drawLine(coords[i], coords[i + 1], coords[i + 2], coords[i + 3], paint);
            canvas.drawLine(coords[i + 2], coords[i + 3], coords[i + 4], coords[i + 5], paint);
            canvas.drawLine(coords[i], coords[i + 1], coords[i + 4], coords[i + 5], paint);
        }
        
        /*
         * float[] coords = {0, 0, 400, 0, 200, 800, 600, 600}; Matrix matrix = new Matrix(); matrix.preRotate(-20); matrix.preTranslate(100, 200); matrix.mapPoints(coords);
         * 
         * canvas.translate(0, 400); canvas.rotate(15);
         * 
         * int gridCount = 1;
         * 
         * paint.setStrokeWidth(1); paint.setColor(Color.BLUE); canvas.drawCircle(coords[0], coords[1], 20, paint); paint.setColor(Color.RED); canvas.drawCircle(coords[2], coords[3], 20, paint);
         * paint.setColor(Color.GREEN); canvas.drawCircle(coords[4], coords[5], 20, paint); paint.setColor(Color.YELLOW); canvas.drawCircle(coords[6], coords[7], 20, paint);
         * paint.setColor(Color.BLUE); canvas.drawLine(coords[0], coords[1], coords[2], coords[3], paint); canvas.drawLine(coords[0], coords[1], coords[4], coords[5], paint);
         * canvas.drawLine(coords[6], coords[7], coords[4], coords[5], paint); canvas.drawLine(coords[6], coords[7], coords[2], coords[3], paint);
         * 
         * // create bitmap from embedded image Bitmap bitmap = BitmapFactory.decodeResource(getContext().getResources(), R.drawable.skull);
         * 
         * // draw the bitmap on the grid canvas.drawBitmapMesh(bitmap, gridCount, gridCount, coords, 0, null, 0, paint);
         */
    }
    
    
    
    @SuppressLint("NewApi")
    private void testDrawVerticesTriangleFanMode(Canvas canvas)
    {
        Paint paint = new Paint();
        
        // draw stuff in black
        paint.setColor(Color.BLACK);
        paint.setStyle(Style.STROKE);
        
        /*
         * Test of drawVertices function - draws a mesh of triangles, shaded by supplied color array
         */
        
        int pointCount = 6;
        
        float[] coords = new float[2 * pointCount];
        // also supply a color array for use with drawVertices
        int[] colors = new int[2 * pointCount];
        
        coords[0] = 1;
        coords[1] = 1;
        
        coords[2] = 2;
        coords[3] = 2;
        
        coords[4] = 2;
        coords[5] = 3;
        
        coords[6] = 1;
        coords[7] = 3;
        
        coords[8] = 0;
        coords[9] = 3;
        
        coords[10] = 0;
        coords[11] = 2;
        
        colors[0] = Color.RED;
        colors[1] = Color.BLUE;
        colors[2] = Color.BLUE;
        
        colors[3] = Color.BLUE;
        colors[4] = Color.BLUE;
        colors[5] = Color.BLUE;
        
        colors[6] = Color.RED;
        colors[7] = Color.WHITE;
        colors[8] = Color.BLUE;
        
        colors[9] = Color.RED;
        colors[10] = Color.WHITE;
        colors[11] = Color.BLUE;
        
        // now draw these on the canvas... but first translate and scale
        canvas.scale(300, 300);
        canvas.translate(2, 1);
        
        // need to disable hardware acceleration, or drawVertices doesn't work
        setLayerType(View.LAYER_TYPE_SOFTWARE, null);
        canvas.drawVertices(Canvas.VertexMode.TRIANGLE_FAN, 12, coords, 0, null, 0, colors, 0, null, 0, 0, paint);
        
        // draw the grid over top for emphasis...
        for (int i = 2; i < coords.length; i += 2)
        {
            canvas.drawLine(coords[0], coords[1], coords[i], coords[i + 1], paint);
            canvas.drawLine(coords[i - 2], coords[i - 1], coords[i], coords[i + 1], paint);
        }
        
        /*
         * float[] coords = {0, 0, 400, 0, 200, 800, 600, 600}; Matrix matrix = new Matrix(); matrix.preRotate(-20); matrix.preTranslate(100, 200); matrix.mapPoints(coords);
         * 
         * canvas.translate(0, 400); canvas.rotate(15);
         * 
         * int gridCount = 1;
         * 
         * paint.setStrokeWidth(1); paint.setColor(Color.BLUE); canvas.drawCircle(coords[0], coords[1], 20, paint); paint.setColor(Color.RED); canvas.drawCircle(coords[2], coords[3], 20, paint);
         * paint.setColor(Color.GREEN); canvas.drawCircle(coords[4], coords[5], 20, paint); paint.setColor(Color.YELLOW); canvas.drawCircle(coords[6], coords[7], 20, paint);
         * paint.setColor(Color.BLUE); canvas.drawLine(coords[0], coords[1], coords[2], coords[3], paint); canvas.drawLine(coords[0], coords[1], coords[4], coords[5], paint);
         * canvas.drawLine(coords[6], coords[7], coords[4], coords[5], paint); canvas.drawLine(coords[6], coords[7], coords[2], coords[3], paint);
         * 
         * // create bitmap from embedded image Bitmap bitmap = BitmapFactory.decodeResource(getContext().getResources(), R.drawable.skull);
         * 
         * // draw the bitmap on the grid canvas.drawBitmapMesh(bitmap, gridCount, gridCount, coords, 0, null, 0, paint);
         */
    }
    
    
    
    @SuppressLint("NewApi")
    private void testDrawVerticesTriangleStripMode(Canvas canvas)
    {
        Paint paint = new Paint();
        
        // draw stuff in black
        paint.setColor(Color.BLACK);
        paint.setStyle(Style.STROKE);
        
        /*
         * Test of drawVertices function - draws a mesh of triangles, shaded by supplied color array
         */
        
        int pointCount = 6;
        
        float[] coords = new float[2 * pointCount];
        // also supply a color array for use with drawVertices
        int[] colors = new int[2 * pointCount];
        
        coords[0] = 1;
        coords[1] = 1;
        
        coords[2] = 2;
        coords[3] = 1;
        
        coords[4] = 2;
        coords[5] = 2;
        
        coords[6] = 3;
        coords[7] = 1;
        
        coords[8] = 3;
        coords[9] = 2;
        
        coords[10] = 4;
        coords[11] = 0;
        
        colors[0] = Color.RED;
        colors[1] = Color.WHITE;
        colors[2] = Color.BLUE;
        
        colors[3] = Color.RED;
        colors[4] = Color.WHITE;
        colors[5] = Color.BLUE;
        
        colors[6] = Color.RED;
        colors[7] = Color.WHITE;
        colors[8] = Color.BLUE;
        
        colors[9] = Color.RED;
        colors[10] = Color.WHITE;
        colors[11] = Color.BLUE;
        
        // now draw these on the canvas... but first translate and scale
        canvas.scale(300, 300);
        canvas.translate(0, 1);
        
        // need to disable hardware acceleration, or drawVertices doesn't work
        setLayerType(View.LAYER_TYPE_SOFTWARE, null);
        canvas.drawVertices(Canvas.VertexMode.TRIANGLE_STRIP, 12, coords, 0, null, 0, colors, 0, null, 0, 0, paint);
        
        // draw the grid over top for emphasis...
        for (int i = 2; i < coords.length; i += 2)
        {
            canvas.drawLine(coords[i - 2], coords[i - 1], coords[i], coords[i + 1], paint);
            if (i > 2)
                canvas.drawLine(coords[i - 4], coords[i - 3], coords[i], coords[i + 1], paint);
        }
        
    }
    
    
    
    private void testText(Canvas canvas)
    {
        // do some stuff
        Paint paint = new Paint();
        
        paint.setTypeface(Typeface.SERIF);
        paint.setColor(Color.RED);
        paint.setTextSize(38);
        paint.setStyle(Style.STROKE);
        
        canvas.drawLine(0, 0, 150, 250, paint);
        canvas.drawText("Some 38-point red text at position 150, 250", 150, 250, paint);
        
        paint.setTextSize(48);
        paint.setColor(Color.GREEN);
        canvas.drawLine(0, 0, 180, 350, paint);
        canvas.drawText("Some 48-point green text at position 180, 350", 180, 350, paint);
        
        // try drawing characters at indicated positions
        float[] pos = { 400, 500, 440, 510, 480, 525, 520, 545, 560, 570, 600, 600, 640, 635, 680, 675, 720, 720, 760, 770, 800, 825};
        String text = "drawPosText";
        canvas.drawPosText(text, pos, paint);
        
        // draw characters along path
        Path path = new Path();
        path.addArc(new RectF(100, 1000, 700, 1400), 180, -180);
        canvas.drawArc(new RectF(100, 1000, 700, 1400), 180, -180, false, paint);
        path.arcTo(new RectF(700, 1000, 1400, 1400), 180, 180);
        canvas.drawArc(new RectF(700, 1000, 1400, 1400), 180, 180, false, paint);
        path.arcTo(new RectF(1400, 1100, 1600, 1300), 180, -180);
        canvas.drawArc(new RectF(1400, 1100, 1600, 1300), 180, -180, false, paint);
        text = "Example of drawTextOnPath with offset and composite path";
        canvas.drawTextOnPath(text, path, 100, 40, paint);
        
    }
    
    
    
    private void testGraphics(Canvas canvas)
    {
        // do some stuff
        Paint paint = new Paint();
        
        // draw some stuff in blue
        paint.setColor(Color.BLUE);
        paint.setStrokeWidth(5);
        paint.setStyle(Style.STROKE);
        
        canvas.drawLine(0, 0, 800, 800, paint);
        canvas.drawCircle(800, 800, 170, paint);
        canvas.drawCircle(800, 800, 10, paint);
        canvas.drawArc(new RectF(750, 750, 850, 850), 180, -90, false, paint);
        
        // fill a rectangle
        paint.setColor(Color.RED);
        paint.setStrokeWidth(3);
        paint.setStyle(Style.FILL);
        canvas.drawLine(0, 0, 1000, 500, paint);
        canvas.drawRect(1000, 500, 1150, 650, paint);
        
        // draw a rounded rectangle
        paint.setStyle(Style.STROKE);
        canvas.drawLine(0, 0, 20, 200, paint);
        canvas.drawRoundRect(new RectF(20, 200, 150, 300), 20, 10, paint);
        
        // draw an oval
        paint.setColor(Color.MAGENTA);
        canvas.drawLine(0, 0, 300, 1000, paint);
        canvas.drawOval(new RectF(200, 800, 400, 1200), paint);
        
    }
    
    
    
    private void testTransformations(Canvas canvas)
    {
        // do some stuff
        Paint paint = new Paint();
        
        // draw a rotated rectangle and some text
        paint.setStyle(Style.STROKE);
        paint.setStrokeWidth(3);
        paint.setTypeface(Typeface.SERIF);
        paint.setTextSize(40);
        paint.setColor(Color.GREEN);
        
        // save state to restore
        canvas.save();
        
        for (int i = 0; i < 40; i += 5)
        {
            canvas.rotate(-5); // counterclockwise...
            canvas.drawLine(0, 0, 250, 1550, paint);
            canvas.drawOval(new RectF(100, 1500, 400, 1600), paint);
        }
        
        canvas.restore();
        
        paint.setColor(Color.BLUE);
        // canvas.scale(2, 2);
        canvas.skew(0.3f, 0);
        canvas.drawLine(0, 0, 75, 380, paint);
        canvas.drawRect(75, 380, 700, 520, paint);
        canvas.drawText("Some 40-point blue text skewed at position 80, 450", 80, 450, paint);
        
        // canvas.setMatrix(origMatrix);
        canvas.skew(-0.3f, 0);
        
        paint.setColor(Color.BLUE);
        canvas.skew(0, 0.8f);
        canvas.drawLine(0, 0, 80, 950, paint);
        canvas.drawRect(80, 880, 700, 1020, paint);
        canvas.drawText("Some 40-point blue text skewed at position 80, 950", 80, 950, paint);
        
        canvas.skew(0, -0.8f);
        
        // both skew and rotate and translate
        paint.setColor(Color.RED);
        canvas.translate(100, 300);
        canvas.rotate(-40); // counterclockwise...
        canvas.skew(-0.4f, -0.3f);
        canvas.drawLine(0, 0, 80, 950, paint);
        canvas.drawRect(80, 880, 700, 1020, paint);
        canvas.drawText("Some 40-point red text translated (100,300), rotated -40, and skewed, at position 80, 950", 80, 950, paint);
        
    }
    
    
    
    private void testClipping(Canvas canvas)
    {
        // test out some clipping stuff
        Paint paint = new Paint();
        paint.setStyle(Style.STROKE);
        
        // first save the canvas state
        canvas.save();
        
        // set clip to rotated rectangle intersected with straight rectangle
        canvas.rotate(-40); // counterclockwise...
        RectF rect = new RectF(0, 200, 900, 1200);
        canvas.drawRect(rect, paint);
        canvas.clipRect(rect);
        canvas.rotate(40);
        rect = new RectF(100, 200, 500, 1000);
        canvas.drawRect(rect, paint);
        canvas.clipRect(rect);
        
        // draw some stuff and see if it gets clipped
        paint.setTypeface(Typeface.SERIF);
        paint.setTextSize(80);
        paint.setColor(Color.GREEN);
        canvas.drawLine(0, 0, 150, 250, paint);
        canvas.drawText("Some 80-point green text at position 150, 250", 150, 250, paint);
        
        // now restore the canvas state and see if clip region is restored
        canvas.restore();
        
        paint.setColor(Color.RED);
        paint.setTextSize(20);
        canvas.drawLine(0, 0, 350, 750, paint);
        canvas.drawText("The above green text should be clipped to the intersection of the two rectangles", 350, 750, paint);
        canvas.drawText("This red text should appear since clipping has been removed", 350, 780, paint);
        
    }
    
    /**
     * Helper class for some of the grid-related tests
     * 
     * @author jsevy
     *
     */
    public class Vector3D
    {
        float x;
        float y;
        float z;
        
        
        
        public Vector3D(float x, float y, float z)
        {
            this.x = x;
            this.y = y;
            this.z = z;
        }
        
        
        
        public Vector3D()
        {
            this.x = 0;
            this.y = 0;
            this.z = 0;
        }
        
        
        
        public float dotProduct(Vector3D other)
        {
            return this.x * other.x + this.y * other.y + this.z * other.z;
        }
        
        
        
        public float magnitude()
        {
            return (float) Math.sqrt(this.dotProduct(this));
        }
        
        
        
        public Vector3D scalarProduct(float scale)
        {
            Vector3D result = new Vector3D();
            
            result.x = this.x * scale;
            result.y = this.y * scale;
            result.z = this.z * scale;
            
            return result;
        }
        
        
        
        public Vector3D crossProduct(Vector3D other)
        {
            Vector3D result = new Vector3D();
            
            result.x = this.y * other.z - other.y * this.z;
            result.y = this.x * other.z - other.x * this.z;
            result.z = this.x * other.y - other.x * this.y;
            
            return result;
        }
        
        
        
        public Vector3D sum(Vector3D other)
        {
            Vector3D result = new Vector3D();
            
            result.x = this.x + other.x;
            result.y = this.y + other.y;
            result.z = this.z + other.z;
            
            return result;
        }
    }
}
